"""
本文件实现了VGG基础结构，以及SSD300对于VGG的扩展结构.
"""

from mxnet.gluon import nn
import mxnet as mx
from mxnet import nd


# num, channel, kernel, stride, padding
# num指定后续元组的重复次数
vgg16_config ={'conv':[(2, (64, 3, 1, 1)), 
                       (2, (128, 3, 1, 1)), 
                       (3, (256, 3, 1, 1)), 
                       (3, (512, 3, 1, 1)), 
                       (3, (512, 3, 1, 1))],
               # pool size, stride
               'pool':[(2,2), (2,2), (2,2), (2,2), (2,2)]} 


class VGGBase(nn.Block):
    """
    VGG基础结构，只包含前五个卷积块以及那五个池化层。
    """
    def __init__(self, config, use_batch=False, ceil_mode=True, 
                 init=mx.init.Xavier, **kwargs):
        super(VGGBase, self).__init__(**kwargs)

        self._config = config
        self._use_batch = use_batch
        self._ceil_mode = ceil_mode
        self._base_conv = 'base_conv_blk'
        self._base_pool = 'base_pool'
        # 初始化方法是Xavier
        self._init = init
        
        # 并不将池化层与卷积块打包到一起，而是在forward中再串起来
        for i, conv in enumerate(self._config['conv'], 1):
            setattr(self, self._base_conv+'%d'%i, 
                    self._get_conv_blk(conv, self._use_batch, self._base_conv+'%d'%i))
        for i, pool in enumerate(self._config['pool'], 1):
            setattr(self, self._base_pool+'%d'%i,
                    self._get_pool(pool, self._ceil_mode, self._base_pool+'%d'%i))
    
    # 因为SSD中在VGG中使用了dialtion，所以这边作为一个默认参数给上
    # 以便后续使用。
    def _get_conv(self, c, k, s, p, d=1, prefix=None):
        """
        返回一个卷积层。
        因为SSD中在VGG中使用了dialtion，所以这边作为一个默认参数给上
        以便后续使用。
        """
        conv = nn.Sequential(prefix=prefix)
        conv.add(nn.Conv2D(channels=c, kernel_size=k, strides=s, padding=p, 
                           dilation=d, weight_initializer=self._init(rnd_type='gaussian', factor_type='out', magnitude=2)))
        if self._use_batch:
            conv.add(nn.BatchNorm())
        conv.add(nn.Activation('relu'))
        return conv
    
    def _get_conv_blk(self, blk_config, use_batch, prefix=None):
        # 返回卷积层块
        blk = nn.Sequential(prefix=prefix)
        with blk.name_scope():
            if blk_config[0] > 0:
                (c, k, s, p) = blk_config[1]
                for i in range(1, blk_config[0]+1):
                    blk.add(self._get_conv(c, k, s, p, prefix='conv%d'%i))
            else:
                for i, (c, k, s, p) in enumerate(blk_config[1:], 1):
                    blk.add(self._get_conv(c, k, s, p, prefix='conv%d'%i))
        return blk
    
    def _get_pool(self, pool_config, ceil_mode, prefix=None):
        return nn.MaxPool2D(pool_size=pool_config[0], strides=pool_config[1],
                            ceil_mode=ceil_mode, prefix=prefix)
    
    def forward(self, x):
        for i in range(1, 6):
            x = getattr(self, self._base_conv+'%d'%i)(x)
            x = getattr(self, self._base_pool+'%d'%i)(x)
        return x

# 0, (channel, kernel, stride, padding), ..., (channel, kernel, stride, padding)
feature_config = {'conv':[(0, (256, 1, 1, 0), (512, 3, 2, 1)),
                          (0, (128, 1, 1, 0), (256, 3, 2, 1)),
                          (0, (128, 1, 1, 0), (256, 3, 1, 0)),
                          (0, (128, 1, 1, 0), (256, 3, 1, 0))]}


class VggFeature(VGGBase):
    """
    继承自VGG基础结构，并按SSD300的结构进行配置。
    该网络前向传播返回的是SSD300需要的六个特征图
    """
    def __init__(self, base_config, feature_config, use_batch=False, 
                 ceil_mode=True, init=mx.init.Xavier, **kwargs):
        super(VggFeature, self).__init__(base_config, use_batch=use_batch, 
                                         ceil_mode=ceil_mode, init=init,
                                         **kwargs)
        # 覆盖第五个pool
        setattr(self, self._base_pool+'5', 
                nn.MaxPool2D(pool_size=3, strides=1, padding=1,
                            ceil_mode=self._ceil_mode, prefix=self._base_pool+'5'))
        
        init_scale = mx.nd.array([0.229, 0.224, 0.225]).reshape((1, 3, 1, 1)) * 255
        self.init_scale = self.params.get_constant('init_scale', init_scale)

        self._feature_conv = 'feature_conv_blk'
        self._feature_config = feature_config
        
        # 对应于论文中对conv4_3进行L2正则化时要学习的参数。
        self._conv4_scale = self.params.get('conv4_scale', shape=(1, 512, 1, 1),
                                            init=mx.init.Constant(20))
        
        # 添加两个转为卷积层的全链接层：
        # 第一个是加入了dilation的。
        self.feature_conv_blk6 = self._get_conv(1024, 3, 1, 6, 6)
        self.feature_conv_blk7 = self._get_conv(1024, 1, 1, 0, 1)
        
        # 添加后四个块
        for i, conv in enumerate(self._feature_config['conv'], 8):
            setattr(self, self._feature_conv+'%d'%i, 
                    self._get_conv_blk(conv, self._use_batch, self._feature_conv+'%d'%i))
            
    def forward(self, x):
        # 这边如果不指定的话，会返回位于cpu上的NDArray值
        # 但是如果网络之前已经在gpu上初始化了，这边会发生错误
        # RuntimeError: Parameter 'vggfeature0_init_scale' was not initialized on context cpu(0). 
        # It was only initialized on [gpu(0), gpu(1)].
        x = x * self.init_scale.data(x.context)
        features = []
        for i in range(1, 4):
            x = getattr(self, self._base_conv + str(i))(x)
            x = getattr(self, self._base_pool + str(i))(x)

        x = getattr(self, self._base_conv+str(4))(x)
        # 这边可千万不能覆盖x啊！！！！
        norm = nd.L2Normalization(x, mode='channel', eps=1e-5) * self._conv4_scale.data(x.context)
        features.append(norm)
        
        x = getattr(self, self._base_pool + str(4))(x)
        x = getattr(self, self._base_conv+str(5))(x)
        x = getattr(self, self._base_pool + str(5))(x)
        x = self.feature_conv_blk6(x)
        x = self.feature_conv_blk7(x)
        features.append(x)
        
        for i in range(8, 12):
            x = getattr(self, self._feature_conv+str(i))(x)
            features.append(x)
        
        return features

def get_VGGFeature():
    return VggFeature(vgg16_config, feature_config)