from .basenet import get_VGGFeature
from .box import BoxGenerator
from mxnet.gluon import nn
from mxnet import nd
import mxnet as mx
import time


scales = [[0.1, 0.14], [0.2, 0.27], [0.37, 0.44], 
          [0.54, 0.62], [0.71, 0.79], [0.88, 0.96]]
ratios = [[1,2,0.5], [1,2,3,0.5,1/3], [1,2,3,0.5,1/3], 
          [1,2,3,0.5,1/3], [1,2,0.5], [1,2,0.5]]


class SSD(nn.Block):
    def __init__(self, base_net, scales, ratios, num_cls=20, 
                 init=mx.init.Xavier, **kwargs):
        super(SSD, self).__init__(**kwargs)
        
        assert len(scales) == len(ratios)
        self.base_net = base_net
        self.scales = list(scales)
        self.ratios = list(ratios)
        self.num_cls = num_cls
        self.init = init
        self.cls_pred = 'cls_pred'
        self.offset_pred = 'offset_pred'
        self.dbox = 'dbox'
        
        # 建立类别预测和偏移预测层
        for i, (scale, ratio) in enumerate(zip(self.scales, self.ratios)):
            num_box = len(scale) + len(ratio) - 1
            setattr(self, self.cls_pred+str(i), 
                    self._get_pred(num_box, self.num_cls, mode='cls', prefix='cls%d'%i))
            setattr(self, self.offset_pred+str(i),
                    self._get_pred(num_box, mode='offset', prefix='offset%d'%i))
            setattr(self, self.dbox+str(i),
                    BoxGenerator(scale, ratio))
        

    def _get_pred(self, num_box, num_cls='', mode='cls', prefix=None):
        """
        根据传入的mode，确定是类别预测层还是偏移预测层。
        """
        assert mode in ('cls', 'offset')
        if mode == 'cls':
            channels = (num_cls + 1) * num_box
        else:
            channels = 4 * num_box
        # 输出层不要有激活函数
        blk = nn.Conv2D(channels, kernel_size=3, strides=1, padding=1, 
                        weight_initializer=self.init(rnd_type='gaussian', factor_type='out', magnitude=2), prefix=prefix)
        return blk
        
    def forward(self, x):
        features = self.base_net(x)

        assert len(features) == len(self.scales)
        
        cls_preds = []
        offset_preds = []
        dboxes = []
        for i, feature in enumerate(features):
            dbox = getattr(self, self.dbox+str(i))(feature)
            # # (1, num, 4)
            dboxes.append(dbox)

            # 各特征图上，输出的n,h,w都不相同
            # (b, 4xn, h, w)
            offset_pred = getattr(self, self.offset_pred+str(i))(feature)
            # (b, 4xn, h, w) -> (b, h, w, 4xn) -> (b, -1)
            offset_preds.append(offset_pred.transpose((0,2,3,1)).flatten())

            # (b, (c+1)xn, h, w)
            cls_pred = getattr(self, self.cls_pred+str(i))(feature)
            # (b, (c+1)xn, h, w) -> (b, h, w, (c+1)xn) -> (b, -1)
            cls_preds.append(cls_pred.transpose((0,2,3,1)).flatten())
            
                # (B, num_box, cls+1)
        return (nd.concat(*cls_preds, dim=1).reshape((0,-1,self.num_cls+1)),
                # (B, num_box*4)
                nd.concat(*offset_preds, dim=1),
                # (1, num_box, 4)
                nd.concat(*dboxes, dim=1),)


def get_SSD300():
    """
    no parameter, just call it
    """
    base_net = get_VGGFeature()
    ssd = SSD(base_net, scales, ratios)
    return ssd