from multidet.model import get_label
from mxnet.gluon.loss import SoftmaxCrossEntropyLoss, HuberLoss
from time import time
import mxnet as mx
from mxnet.gluon import utils as gutils
import logging
import argparse
from multidet import data
from multidet import model
from multidet.utils import get_gpu
from mxnet import nd, gluon


def cls_eval(cls_preds, cls_labels, cls_masks):
    a = ((cls_preds.argmax(axis=-1) == cls_labels)*cls_masks).sum().asscalar()
    return a / cls_masks.sum().asscalar()

def bbox_eval(bbox_preds, bbox_labels, bbox_masks):
    return ((bbox_labels - bbox_preds) * bbox_masks).abs().sum().asscalar()


def ssd_train_batch(model, X, y, trainer, loss, batch_size, evals, ctx):
    Xs = gutils.split_and_load(X, ctx_list=ctx)
    ys = gutils.split_and_load(y, ctx_list=ctx)

    with mx.autograd.record():
        # (cls_preds, box_preds, anchors)...
        model_out = [model(xi) for xi in Xs]
        with mx.autograd.pause():
            # (cls_targets, cls_mask, box_targets, box_masks, num_positive)...
            targets = [get_label(out[2], yi, out[0]) \
                       for out, yi in zip(model_out, ys)]

        closses = [loss[0](out[0], target[0], target[1].expand_dims(axis=-1))/target[4] \
                   for out, target in zip(model_out, targets)]

        llosses = [loss[1](out[1], target[2], target[3])/target[4] \
                   for out, target in zip(model_out, targets)]

        ls = [(cl + ll) \
              for cl, ll in zip(closses, llosses)]

        mx.autograd.backward(ls)

    trainer.step(1)

    # 只统计第一个gpu上的loss
    # [number, number, number]，返回后求和，然后除以批次
    batch_loss =  [closses[0].sum().asscalar(), \
                   llosses[0].sum().asscalar()]

    # 只统计第一个gpu上的精度
    # [number, number]，返回后求和，然后除以批次
    batch_acc = [evals[0](model_out[0][0], targets[0][0], \
                          # 过滤掉背景的预测准确率
                          targets[0][0]>0), \
                 evals[1](model_out[0][1], targets[0][2], \
                          targets[0][3])]

    return batch_loss, batch_acc


SSDloss = [SoftmaxCrossEntropyLoss(), HuberLoss()]
SSDeval = [cls_eval, bbox_eval]


def model_train(model, train_iter, num_epoch, trainer, batch_size,
                loss, evals, val_iter=None, ctx=None, logger=None):

    for epoch in range(num_epoch):
        cls_loss, offset_loss = 0.0, 0.0
        cls_acc, box_acc = 0.0, 0.0

        start = time()
        i = 0

        for X, y in train_iter:
            batch_start = time()
            i += 1
            batch_loss, batch_acc = ssd_train_batch(model, X, y,
                                    trainer, loss, batch_size, evals, ctx)
    
            cls_loss += batch_loss[0]
            offset_loss += batch_loss[1]

            cls_acc += batch_acc[0]
            box_acc += batch_acc[1]
    
            logger.info('[epoch {}][batch {}], time:{:.1f}s'.format(epoch, i, time()-batch_start))
            logger.info('\tcls acc:{:.3e}, box acc:{:.3e}, cls loss:{:.4e}, box loss:{:.4e}'.
                        format(cls_acc/i, box_acc/i, cls_loss/i, offset_loss/i))
            
           
        logger.info('[epoch {}], time:{:.1f}s'.format(epoch, time()-start))
        logger.info('\tcls acc:{:.3e}, box acc:{:.3e}, cls loss:{:.4e}, box loss:{:.4e}'.
                    format(cls_acc/i, box_acc/i, cls_loss/i, offset_loss/i))
        
        logger.info('==========================================================')

    
        model.save_parameters('ssd-params_epoch%d.param'%epoch)

        # for c in ctx:
        #     c.empty_cache()


def parse_args():
    parser = argparse.ArgumentParser(description='Train SSD300.')
    parser.add_argument('idx', help='the path of idx file')
    parser.add_argument('rec', help='the path of rec file')
    parser.add_argument('--pre', dest='pre_params', default=None,
                        help='pretrained weight path')
    parser.add_argument('--base', dest='base_weight', default=None,
                        help='the paht of the weight of vggbase network')
    parser.add_argument('--batch', dest='batch_num', type=int, default=32,
                        help='the data batch number')
    parser.add_argument('--epoch', dest='epoch_num', type=int, default=100,
                        help='the train epoch number')
    parser.add_argument('--lr', dest='lr', type=float, default=1e-3,
                        help='the learning rate')
    parser.add_argument('--wd', dest='wd', type=float, default=0.0005,
                        help='the weight decay')
    parser.add_argument('--momentum', dest='momentum', type=float, default=0.9,
                        help='the momentum for sgd')
    parser.add_argument('--log', dest='log_path', default='./train.log')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    # data
    train_iter = data.DataIter(args.idx, args.rec, batch_size=args.batch_num,
                               shuffle=True, aug_list=data.auglist_full)
    ctx = get_gpu()

    # model
    ssd = model.get_SSD300()
    ssd.initialize(ctx=ctx)

    if args.pre_params is not None:
        ssd.load_parameters(args.pre_params)

    elif args.base_weight is not None:
        preparams = nd.load(args.base_weight)
        for key in preparams.keys():
            ssd.base_net.collect_params()[key].set_data(preparams[key])

    # trainer
    trainer = gluon.Trainer(ssd.collect_params(), 'sgd', 
                            {'learning_rate':args.lr,
                             'wd':args.wd,
                             'momentum':args.momentum})

    # logger
    logging.basicConfig()
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    fh = logging.FileHandler(args.log_path)
    logger.addHandler(fh)
    logger.info('Config Done, begin training. training on {!r}'.format(ctx))

    model_train(ssd, train_iter, args.epoch_num, trainer, args.batch_num,
                SSDloss, SSDeval, val_iter=None, ctx=ctx, logger=logger)




